
# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

import numpy as np

from cherab.core.utility import PhotonToJ
from cherab.core.atomic.elements import lookup_isotope, hydrogen, deuterium, tritium
from cherab.openadas import OpenADAS

from cherab.solps.mesh_geometry import SOLPSMesh
from cherab.solps.solps_plasma import SOLPSSimulation, prefer_element, eirene_flux_to_velocity, b2_flux_to_velocity


def load_solps_from_mdsplus(mds_server, ref_number, atomic_data=None):
    """
    Load a SOLPS simulation from a MDSplus server.

    :param str mds_server: Server address.
    :param int ref_number: Simulation reference number.
    :param AtomicData atomic_data: The atomic data provider. Used here to convert
                                   the radiation density from photons/s/m3 to W/m3.
                                   Defaults to `OpenADAS()` if None.
    :rtype: SOLPSSimulation
    """

    from MDSplus import Connection as MDSConnection, MdsException

    # Setup connection to server
    conn = MDSConnection(mds_server)
    conn.openTree('solps', ref_number)

    # Load SOLPS mesh geometry and lookup arrays
    mesh = load_mesh_from_mdsplus(conn, MdsException)

    # Load each plasma species in simulation
    ns = conn.get(r'\SOLPS::TOP.IDENT.NS').data()  # Number of species
    zn = conn.get(r'\SOLPS::TOP.SNAPSHOT.GRID.ZN').data().astype(int)  # Nuclear charge
    am = np.round(conn.get(r'\SOLPS::TOP.SNAPSHOT.GRID.AM').data()).astype(int)  # Atomic mass number
    charge = conn.get(r'\SOLPS::TOP.SNAPSHOT.GRID.ZA').data().astype(int)   # Ionisation/charge

    species_list = []
    neutral_indx = []
    hydrogen_neutrals = {}
    for i in range(ns):
        isotope = lookup_isotope(zn[i], number=am[i])
        species = prefer_element(isotope)  # Prefer Element over Isotope if the mass number is the same
        species_list.append((species.name, charge[i]))
        if charge[i] == 0:
            neutral_indx.append(i)
            if species in (hydrogen, deuterium, tritium):
                hydrogen_neutrals[species] = i

    sim = SOLPSSimulation(mesh, species_list)
    nx = mesh.nx
    ny = mesh.ny

    ##########################
    # Magnetic field vectors #
    sim.b_field = conn.get(r'\SOLPS::TOP.SNAPSHOT.B').data()[:3]
    # sim.b_field_cylindrical is created automatically

    # Load electron temperature and density
    sim.electron_temperature = conn.get(r'\SOLPS::TOP.SNAPSHOT.TE').data()
    sim.electron_density = conn.get(r'\SOLPS::TOP.SNAPSHOT.NE').data()

    # Load ion temperature
    sim.ion_temperature = conn.get(r'\SOLPS::TOP.SNAPSHOT.TI').data()

    # Load species density
    sim.species_density = conn.get(r'\SOLPS::TOP.SNAPSHOT.NA').data()

    # Load parallel velocity
    parallel_velocity = conn.get(r'\SOLPS::TOP.SNAPSHOT.UA').data()

    # Load poloidal and radial particle fluxes for velocity calculation
    poloidal_flux = conn.get(r'\SOLPS::TOP.SNAPSHOT.FNAX').data()
    radial_flux = conn.get(r'\SOLPS::TOP.SNAPSHOT.FNAY').data()

    # B2 fluxes are defined between cells, so correcting array shapes if needed
    if poloidal_flux.shape[2] == nx - 1:
        poloidal_flux = np.concatenate((np.zeros((ns, ny, 1)), poloidal_flux), axis=2)

    if radial_flux.shape[1] == ny - 1:
        radial_flux = np.concatenate((np.zeros((ns, 1, nx)), radial_flux), axis=1)

    # Setting velocities from B2 flux
    sim.velocities_cylindrical = b2_flux_to_velocity(sim, poloidal_flux, radial_flux, parallel_velocity)

    # Obtaining additional data from EIRENE and replacing data for neutrals

    try:
        # Replace the species densities
        neutral_density = conn.get(r'\SOLPS::TOP.SNAPSHOT.DAB2').data()  # this will throw a TypeError is neutral_density is not an array
        # We can update the data without re-initialising interpolators because they use pointers
        sim.species_density[neutral_indx] = neutral_density[:]

    except (MdsException, TypeError):
        print("Warning! This is B2 stand-alone simulation.")
        b2_standalone = True
    else:
        b2_standalone = False

    if not b2_standalone:
        # Obtaining neutral atom velocity from EIRENE flux
        # Note that if the output for fluxes was turned off, PFLA and RFLA' are all zeros
        try:
            neutral_poloidal_flux = conn.get(r'\SOLPS::TOP.SNAPSHOT.PFLA').data()[:]
            neutral_radial_flux = conn.get(r'\SOLPS::TOP.SNAPSHOT.RFLA').data()[:]

            if np.any(neutral_poloidal_flux) or np.any(neutral_radial_flux):
                sim.velocities_cylindrical[neutral_indx] = eirene_flux_to_velocity(sim, neutral_poloidal_flux, neutral_radial_flux,
                                                                                   parallel_velocity[neutral_indx])
                sim.velocities_cylindrical = sim.velocities_cylindrical  # Updating sim.velocities

        except (MdsException, TypeError):
            pass

        # Obtaining neutral temperatures
        try:
            sim.neutral_temperature = conn.get(r'\SOLPS::TOP.SNAPSHOT.TAB2').data()[:]
        except (MdsException, TypeError):
            pass

    ###############################
    # Load extra data from server #
    ###############################

    ####################
    # Integrated power #
    try:
        linerad = np.sum(conn.get(r'\SOLPS::TOP.SNAPSHOT.RQRAD').data()[:], axis=0)
    except (MdsException, TypeError):
        linerad = 0

    try:
        brmrad = np.sum(conn.get(r'\SOLPS::TOP.SNAPSHOT.RQBRM').data()[:], axis=0)
    except (MdsException, TypeError):
        brmrad = 0

    try:
        eneutrad = conn.get(r'\SOLPS::TOP.SNAPSHOT.ENEUTRAD').data()[:]
        if np.ndim(eneutrad) == 3:
            neurad = np.abs(np.sum(eneutrad, axis=0))
        else:
            neurad = np.abs(eneutrad)
    except (MdsException, TypeError):
        neurad = 0

    total_rad = linerad + brmrad + neurad

    if np.any(total_rad != 0):
        sim.total_radiation = total_rad / mesh.vol

    ########################################
    # Molecular and total H-alpha emission #
    if len(hydrogen_neutrals):
        try:
            halpha_mol = conn.get(r'\SOLPS::TOP.SNAPSHOT.EMISSMOL').data()[:]
        except (MdsException, TypeError):
            halpha_mol = 0

        try:
            halpha_at = conn.get(r'\SOLPS::TOP.SNAPSHOT.EMISS').data()[:]
        except (MdsException, TypeError):
            halpha_at = 0

        halpha_total = halpha_mol + halpha_at

        if isinstance(halpha_total, np.ndarray):
            atomic_data = atomic_data or OpenADAS()
            total_hydrogen_density = sim.species_density[list(hydrogen_neutrals.values())].sum(0)
            effective_energy = 0
            for isotope, i in hydrogen_neutrals.items():
                wavelength = atomic_data.wavelength(isotope, 0, (3, 2))
                fraction = np.divide(sim.species_density[i], total_hydrogen_density,
                                     out=np.zeros_like(total_hydrogen_density),
                                     where=(total_hydrogen_density > 0))
                effective_energy += PhotonToJ.to(fraction, wavelength)
            sim.halpha_mol_radiation = halpha_mol * effective_energy  # photon s-1 m-3 --> W m-3
            sim.halpha_total_radiation = halpha_total * effective_energy  # photon s-1 m-3 --> W m-3

    return sim


def load_mesh_from_mdsplus(mds_connection, MdsException):
    """
    Load the SOLPS mesh geometry for a given MDSplus connection.

    :param mds_connection: MDSplus connection object. Already set to the SOLPS tree with pulse #ID.
    :param mdsExceptions: MDSplus mdsExceptions module for error handling.
    """

    # Load the R, Z coordinates of the cell vertices, original coordinates are (4, 38, 98)
    r = mds_connection.get(r'\TOP.SNAPSHOT.GRID:R').data()
    z = mds_connection.get(r'\TOP.SNAPSHOT.GRID:Z').data()

    vol = mds_connection.get(r'\SOLPS::TOP.SNAPSHOT.VOL').data()

    # Loading neighbouring cell indices
    neighbix = np.zeros(r.shape, dtype=int)
    neighbiy = np.zeros(r.shape, dtype=int)

    neighbix[0] = mds_connection.get(r'\SOLPS::TOP.SNAPSHOT.GRID:LEFTIX').data().astype(int)
    neighbix[1] = mds_connection.get(r'\SOLPS::TOP.SNAPSHOT.GRID:BOTTOMIX').data().astype(int)
    neighbix[2] = mds_connection.get(r'\SOLPS::TOP.SNAPSHOT.GRID:RIGHTIX').data().astype(int)
    neighbix[3] = mds_connection.get(r'\SOLPS::TOP.SNAPSHOT.GRID:TOPIX').data().astype(int)

    neighbiy[0] = mds_connection.get(r'\SOLPS::TOP.SNAPSHOT.GRID:LEFTIY').data().astype(int)
    neighbiy[1] = mds_connection.get(r'\SOLPS::TOP.SNAPSHOT.GRID:BOTTOMIY').data().astype(int)
    neighbiy[2] = mds_connection.get(r'\SOLPS::TOP.SNAPSHOT.GRID:RIGHTIY').data().astype(int)
    neighbiy[3] = mds_connection.get(r'\SOLPS::TOP.SNAPSHOT.GRID:TOPIY').data().astype(int)

    neighbix[neighbix == r.shape[2]] = -1
    neighbiy[neighbiy == r.shape[1]] = -1

    # build mesh object
    mesh = SOLPSMesh(r, z, vol, neighbix, neighbiy)

    #############################
    # Add additional parameters #
    #############################

    # add the vessel geometry
    try:
        vessel = mds_connection.get(r'\SOLPS::TOP.SNAPSHOT.GRID:VESSEL').data()[:]
        mesh.vessel = vessel
    except (MdsException, TypeError, IndexError):
        pass

    return mesh



# Copyright 2016-2021 Euratom
# Copyright 2016-2021 United Kingdom Atomic Energy Authority
# Copyright 2016-2021 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

from raysect.core.math.function.float cimport Function3D
from cherab.core.atomic cimport Line
from cherab.core.plasma cimport PlasmaModel
from cherab.core.species cimport Species
from cherab.core.model.lineshape cimport LineShapeModel


cdef class SOLPSLineEmitter(PlasmaModel):

    cdef:
        Line _line
        double _wavelength
        Species _target_species
        LineShapeModel _lineshape
        Function3D _power_density
        
        object _lineshape_class, _lineshape_args, _lineshape_kwargs

    cdef int _populate_cache(self) except -1




# Copyright 2016-2021 Euratom
# Copyright 2016-2021 United Kingdom Atomic Energy Authority
# Copyright 2016-2021 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

from raysect.core import translate
from raysect.optical.material import VolumeTransform
from raysect.primitive import Cylinder, Subtract

from cherab.core.math.mappers import AxisymmetricMapper
from cherab.tools.emitters import RadiationFunction

from cherab.solps import SOLPSMesh, SOLPSFunction2D


def make_solps_emitter(solps_mesh, radiation_function, parent=None, step=0.01):
    """
    Non-spectral emitter with the emissivity defined as SOLPSFunction2D.

    :param SOLPSMesh solps_mesh: SOLPS simulation mesh.
    :param SOLPSFunction2D radiation_function: Emissivity in W m-3.
    :param Node parent: parent node in the scenegraph, e.g. a World object.
    :param float step: Volume integration step in meters.

    :rtype: Primitive
    """

    if not isinstance(solps_mesh, SOLPSMesh):
        raise TypeError('Argument solps_mesh must be a SOLPSMesh instance.')
    if not isinstance(radiation_function, SOLPSFunction2D):
        raise TypeError('Argument radiation_function must be a SOLPSFunction2D instance.')

    radiation_function_3d = AxisymmetricMapper(radiation_function)
    outer_radius = solps_mesh.mesh_extent['maxr']
    inner_radius = solps_mesh.mesh_extent['minr']
    height = solps_mesh.mesh_extent['maxz'] - solps_mesh.mesh_extent['minz']
    lower_z = solps_mesh.mesh_extent['minz']
    emitter = RadiationFunction(radiation_function_3d, step=step)
    material = VolumeTransform(emitter, transform=translate(0, 0, -lower_z))

    plasma_volume = Subtract(Cylinder(outer_radius, height), Cylinder(inner_radius, height),
                             material=material, parent=parent, transform=translate(0, 0, lower_z))

    return plasma_volume



# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

from cherab.solps.eirene.parser.fort44_post_2016 import load_fort44_post_2016
from cherab.solps.eirene.parser.fort44_2013 import load_fort44_2013
from cherab.solps.eirene.parser.fort44_pre_2007 import load_fort44_pre_2007


def load_fort44_file(file_path, debug=False):
    """ Read neutral species and wall flux information from fort.44

    Template for reading is ngread.F in b2plot of B2.5 source.

    Specifications of data format are in Section 5.2 of the SOLPS
    manual (Running the coupled version -> Output Files).

    :param str file_path: path to EIRENE "fort.44" output file
    :param bool debug: status flag for printing debugging output
    :rtype:
    """
    with open(file_path, 'r') as file_handle:
        # dictionary with file data
        data = {}

        # Read sizes
        line = file_handle.readline().split()
        data["nx"] = int(line[0])
        data["ny"] = int(line[1])
        data["version"] = int(line[2])

    if debug:
        print('Geometry & Version : nx {}, ny {}, version {}'
              .format(data["nx"], data["ny"], data["version"]))

        # Look up file parsing function and call it to obtain for44 block and update data dictionary
    parser = assign_fort44_parser(data["version"])
    return parser(file_path, debug)


def assign_fort44_parser(file_version):
    """
    Looks up file parsing function for the parser Eirene file.
    :param file_version: Fort44 file version from the file header.
    :return: Parsing function object
    """

    # TODO write universal fort44 parser
    file_version = int(file_version)
    # Parser load_fort44_pre_2007 is used also for 20071209, 20080706 and 20081111
    # versions as a temporary solution. Parameter 'eneutrad' is not parsed from these file versions.
    if file_version <= 20081111:
        return load_fort44_pre_2007
    if file_version == 20130210:
        return load_fort44_2013
    if file_version >= 20160829:
        return load_fort44_post_2016
    raise ValueError("Can't read version {} fort.44 file".format(file_version))




# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

import numpy as np
from scipy.io import netcdf
from scipy import constants

from cherab.core.atomic.elements import lookup_isotope

from cherab.solps.mesh_geometry import SOLPSMesh
from cherab.solps.solps_plasma import SOLPSSimulation, prefer_element, eirene_flux_to_velocity, b2_flux_to_velocity


def load_solps_from_balance(balance_filename):
    """
    Load a SOLPS simulation from SOLPS balance.nc output files.
    """

    el_charge = constants.elementary_charge
    rydberg_energy = constants.value('Rydberg constant times hc in eV')

    # Open the file
    with netcdf.netcdf_file(balance_filename, 'r') as fhandle:

        # Load SOLPS mesh geometry
        mesh = load_mesh_from_netcdf(fhandle)

        # Load each plasma species in simulation

        species_list = []
        neutral_indx = []
        am = np.round(fhandle.variables['am'].data).astype(int)  # Atomic mass number
        charge = fhandle.variables['za'].data.astype(int)   # Ionisation/charge
        species_names = fhandle.variables['species'].data.copy()
        ns = am.size
        for i in range(ns):
            symbol = ''.join([b.decode('utf-8').strip(' 0123456789+-') for b in species_names[i]])  # also strips isotope number
            if symbol not in ('D', 'T'):
                isotope = lookup_isotope(symbol, number=am[i])  # will throw an error for D or T
                species = prefer_element(isotope)  # Prefer Element over Isotope if the mass number is the same
            else:
                species = lookup_isotope(symbol)

            # If we only need to populate species_list, there is probably a faster way to do this...
            species_list.append((species.name, charge[i]))
            if charge[i] == 0:
                neutral_indx.append(i)

        sim = SOLPSSimulation(mesh, species_list)
        nx = mesh.nx
        ny = mesh.ny

        ##########################
        # Magnetic field vectors #
        sim.b_field = fhandle.variables['bb'].data.copy()[:3]
        # sim.b_field_cylindrical is created automatically

        # Load electron species
        sim.electron_temperature = fhandle.variables['te'].data.copy() / el_charge
        sim.electron_density = fhandle.variables['ne'].data.copy()

        # Load ion temperature
        sim.ion_temperature = fhandle.variables['ti'].data.copy() / el_charge

        # Load species density
        sim.species_density = fhandle.variables['na'].data.copy()

        # Load parallel velocity
        parallel_velocity = fhandle.variables['ua'].data.copy()

        # Load poloidal and radial particle fluxes for velocity calculation
        if 'fna_tot' in fhandle.variables:
            fna = fhandle.variables['fna_tot'].data.copy()
            # Obtaining velocity from B2 flux
            sim.velocities_cylindrical = b2_flux_to_velocity(sim, fna[:, 0], fna[:, 1], parallel_velocity)
        else:  # trying to obtain particle flux from components
            fna = 0
            for key in fhandle.variables.keys():
                if 'fna_' in key:
                    fna += fhandle.variables[key].data.copy()
            if np.any(fna != 0):
                # Obtaining velocity from B2 flux
                sim.velocities_cylindrical = b2_flux_to_velocity(sim, fna[:, 0], fna[:, 1], parallel_velocity)

        # Obtaining additional data from EIRENE and replacing data for neutrals
        if 'dab2' in fhandle.variables:
            sim.species_density[neutral_indx] = fhandle.variables['dab2'].data.copy()[:, :ny, :nx]  # in case of large grid size
            b2_standalone = False
        else:
            b2_standalone = True

        if not b2_standalone:

            # Obtaining neutral atom velocity from EIRENE flux
            # Note that if the output for fluxes was turned off, pfluxa and rfluxa' are all zeros
            if 'pfluxa' in fhandle.variables and 'rfluxa' in fhandle.variables:
                neutral_poloidal_flux = fhandle.variables['pfluxa'].data.copy()[:, :ny, :nx]
                neutral_radial_flux = fhandle.variables['rfluxa'].data.copy()[:, :ny, :nx]

                if np.any(neutral_poloidal_flux) or np.any(neutral_radial_flux):
                    neutral_velocities = eirene_flux_to_velocity(sim, neutral_poloidal_flux, neutral_radial_flux,
                                                                 parallel_velocity[neutral_indx])
                    if sim.velocities_cylindrical is not None:
                        sim.velocities_cylindrical[neutral_indx] = neutral_velocities
                        sim.velocities_cylindrical = sim.velocities_cylindrical  # Updating sim.velocities
                    else:
                        # No 'fna_*' keys in balance.nc and b2 species velocities are not set
                        velocities_cylindrical = np.zeros((len(sim.species_list, 3, ny, nx)))
                        velocities_cylindrical[neutral_indx] = neutral_velocities
                        sim.velocities_cylindrical = velocities_cylindrical

            # Obtaining neutral temperatures
            if 'tab2' in fhandle.variables:
                sim.neutral_temperature = fhandle.variables['tab2'].data.copy()[:, :ny, :nx]

            # Calculate the total radiated power
            b2_ploss = 0
            eirene_ecoolrate = 0
            eirene_potential_loss = 0

            # Total radiated power from B2, not including neutrals
            if 'b2stel_she_bal' in fhandle.variables:
                b2_ploss = np.sum(fhandle.variables['b2stel_she_bal'].data, axis=0) / mesh.vol

            # Electron energy loss due to interactions with neutrals
            if 'eirene_mc_eael_she_bal' in fhandle.variables:
                eirene_ecoolrate = np.sum(fhandle.variables['eirene_mc_eael_she_bal'].data, axis=0) / mesh.vol

            # Ionisation rate from EIRENE, needed to calculate the energy loss to overcome the ionisation potential of atoms
            if 'eirene_mc_papl_sna_bal' in fhandle.variables:
                tmp = np.sum(fhandle.variables['eirene_mc_papl_sna_bal'].data, axis=0)[1]
                eirene_potential_loss = rydberg_energy * tmp * el_charge / mesh.vol

            # This will be negative (energy sink); multiply by -1
            total_rad = -1.0 * (b2_ploss + (eirene_ecoolrate - eirene_potential_loss))

        else:
            # Total radiated power from B2, not including neutrals
            b2_ploss = 0
            potential_loss = 0

            if 'b2stel_she_bal' in fhandle.variables:
                b2_ploss = np.sum(fhandle.variables['b2stel_she_bal'].data, axis=0) / mesh.vol

            if 'b2stel_sna_ion_bal' in fhandle.variables:
                potential_loss = np.sum(fhandle.variables['b2stel_sna_ion_bal'].data, axis=0) / mesh.vol

            # Save total radiated power to the simulation object
            total_rad = rydberg_energy * el_charge * potential_loss - b2_ploss

        if np.any(total_rad != 0):
            sim.total_radiation = total_rad

    return sim


def load_mesh_from_netcdf(fhandle):

    # Load SOLPS mesh geometry
    # Re-arrange the array dimensions in the way CHERAB expects...
    r = fhandle.variables['crx'].data.copy()
    z = fhandle.variables['cry'].data.copy()
    vol = fhandle.variables['vol'].data.copy()

    # Loading neighbouring cell indices
    neighbix = np.zeros(r.shape, dtype=int)
    neighbiy = np.zeros(r.shape, dtype=int)

    neighbix[0] = fhandle.variables['leftix'].data.astype(int)  # poloidal prev.
    neighbix[1] = fhandle.variables['bottomix'].data.astype(int)  # radial prev.
    neighbix[2] = fhandle.variables['rightix'].data.astype(int)  # poloidal next
    neighbix[3] = fhandle.variables['topix'].data.astype(int)  # radial next

    neighbiy[0] = fhandle.variables['leftiy'].data.astype(int)
    neighbiy[1] = fhandle.variables['bottomiy'].data.astype(int)
    neighbiy[2] = fhandle.variables['rightiy'].data.astype(int)
    neighbiy[3] = fhandle.variables['topiy'].data.astype(int)

    # In SOLPS cell indexing starts with -1 (guarding cell), but in SOLPSMesh -1 means no neighbour.
    neighbix += 1
    neighbiy += 1
    neighbix[neighbix == r.shape[2]] = -1
    neighbiy[neighbiy == r.shape[1]] = -1

    # Create the SOLPS mesh
    mesh = SOLPSMesh(r, z, vol, neighbix, neighbiy)

    return mesh



# cython: language_level=3

# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

import numpy as np
cimport numpy as np

from raysect.core.math.function.float.function2d.interpolate.common cimport MeshKDTree2D
from raysect.core.math.vector cimport Vector3D, new_vector3d
from raysect.core.math.point cimport new_point2d

from cherab.core.math.function cimport Function2D, VectorFunction2D
cimport cython


cdef class SOLPSFunction2D(Function2D):

    def __init__(self, object vertex_coords not None, object triangles not None, object triangle_to_grid_map not None, object grid_data not None):

        # use numpy arrays to store data internally
        vertex_coords = np.array(vertex_coords, dtype=np.float64)
        triangles = np.array(triangles, dtype=np.int32)
        triangle_to_grid_map = np.array(triangle_to_grid_map, dtype=np.int32)

        # Attention!!! Do not copy grid_data! Attribute self._grid_data must point to the original data array,
        # so as not to re-initialize the interpolator if the user changes data values.

        # build kdtree
        self._kdtree = MeshKDTree2D(vertex_coords, triangles)

        # populate internal attributes
        self._grid_data = grid_data
        self._triangle_to_grid_map = triangle_to_grid_map

        self._grid_data_mv = self._grid_data
        self._triangle_to_grid_map_mv = self._triangle_to_grid_map

    def __getstate__(self):
        return self._grid_data, self._triangle_to_grid_map, self._kdtree

    def __setstate__(self, state):
        self._grid_data, self._triangle_to_grid_map, self._kdtree = state
        self._triangle_to_grid_map_mv = self._triangle_to_grid_map
        self._grid_data_mv = self._grid_data

    def __reduce__(self):
        return self.__new__, (self.__class__, ), self.__getstate__()

    @classmethod
    def instance(cls, SOLPSFunction2D instance not None, object grid_data=None):
        """
        Creates a new interpolator instance from an existing interpolator instance.
        The new interpolator instance will share the same internal acceleration
        data as the original interpolator. The grid_data of the new instance can
        be redefined.
        This method should be used if the user has multiple sets of grid_data
        that lie on the same mesh geometry. Using this methods avoids the
        repeated rebuilding of the mesh acceleration structures by sharing the
        geometry data between multiple interpolator objects.
        :param SOLPSFunction2D instance: SOLPSFunction2D object.
        :param ndarray grid_data: An array containing data on SOLPS grid.
        :return: A SOLPSFunction2D object.
        :rtype: SOLPSFunction2D
        """

        cdef SOLPSFunction2D m

        # copy source data
        m = SOLPSFunction2D.__new__(SOLPSFunction2D)
        m._kdtree = instance._kdtree
        m._triangle_to_grid_map = instance._triangle_to_grid_map

        # do we have replacement triangle data?
        if grid_data is None:
            m._grid_data = instance._grid_data
        else:
            m._grid_data = grid_data

        m._triangle_to_grid_map_mv = m._triangle_to_grid_map
        m._grid_data_mv = m._grid_data

        return m

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef double evaluate(self, double x, double y) except? -1e999:

        cdef:
            np.int32_t triangle_id, ix, iy

        if self._kdtree.is_contained(new_point2d(x, y)):

            triangle_id = self._kdtree.triangle_id
            ix = self._triangle_to_grid_map_mv[triangle_id, 0]
            iy = self._triangle_to_grid_map_mv[triangle_id, 1]
            return self._grid_data_mv[ix, iy]

        return 0.0

cdef class SOLPSVectorFunction2D(VectorFunction2D):

    def __init__(self, object vertex_coords not None, object triangles not None, object triangle_to_grid_map not None, object grid_vectors not None):

        # use numpy arrays to store data internally
        vertex_coords = np.array(vertex_coords, dtype=np.float64)
        triangles = np.array(triangles, dtype=np.int32)
        triangle_to_grid_map = np.array(triangle_to_grid_map, dtype=np.int32)

        # Attention!!! Do not copy grid_vectors! Attribute self._grid_vectors must point to the original data array,
        # so as not to re-initialize the interpolator if the user changes data values.

        # build kdtree
        self._kdtree = MeshKDTree2D(vertex_coords, triangles)

        # populate internal attributes
        self._grid_vectors = grid_vectors
        self._triangle_to_grid_map = triangle_to_grid_map
        self._grid_vectors_mv = self._grid_vectors
        self._triangle_to_grid_map_mv = self._triangle_to_grid_map

    def __getstate__(self):
        return self._grid_vectors, self._triangle_to_grid_map, self._kdtree

    def __setstate__(self, state):
        self._grid_vectors, self._triangle_to_grid_map, self._kdtree = state
        self._grid_vectors_mv = self._grid_vectors
        self._triangle_to_grid_map_mv = self._triangle_to_grid_map

    def __reduce__(self):
        return self.__new__, (self.__class__, ), self.__getstate__()

    @classmethod
    def instance(cls, SOLPSVectorFunction2D instance not None, object grid_vectors=None):
        """
        Creates a new interpolator instance from an existing interpolator instance.
        The new interpolator instance will share the same internal acceleration
        data as the original interpolator. The grid_data of the new instance can
        be redefined.
        This method should be used if the user has multiple sets of grid_data
        that lie on the same mesh geometry. Using this methods avoids the
        repeated rebuilding of the mesh acceleration structures by sharing the
        geometry data between multiple interpolator objects.
        :param SOLPSVectorFunction2D instance: SOLPSVectorFunction2D object.
        :param ndarray grid_vectors: An array containing vector data on SOLPS grid.
        :return: A SOLPSVectorFunction2D object.
        :rtype: SOLPSVectorFunction2D
        """

        cdef SOLPSVectorFunction2D m

        # copy source data
        m = SOLPSVectorFunction2D.__new__(SOLPSVectorFunction2D)
        m._kdtree = instance._kdtree
        m._triangle_to_grid_map = instance._triangle_to_grid_map

        # do we have replacement triangle data?
        if grid_vectors is None:
            m._grid_vectors = instance._grid_vectors
        else:
            m._grid_vectors = grid_vectors

        m._triangle_to_grid_map_mv = m._triangle_to_grid_map
        m._grid_vectors_mv = m._grid_vectors

        return m

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef Vector3D evaluate(self, double x, double y):

        cdef:
            np.int32_t triangle_id, ix, iy
            double vx, vy, vz

        if self._kdtree.is_contained(new_point2d(x, y)):

            triangle_id = self._kdtree.triangle_id
            ix = self._triangle_to_grid_map_mv[triangle_id, 0]
            iy = self._triangle_to_grid_map_mv[triangle_id, 1]
            vx = self._grid_vectors_mv[0, ix, iy]
            vy = self._grid_vectors_mv[1, ix, iy]
            vz = self._grid_vectors_mv[2, ix, iy]

            return new_vector3d(vx, vy, vz)

        return new_vector3d(0, 0, 0)




# Copyright 2016-2021 Euratom
# Copyright 2016-2021 United Kingdom Atomic Energy Authority
# Copyright 2016-2021 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.collections import PolyCollection


class SOLPSMesh:
    """
    SOLPSMesh geometry object.

    The SOLPS mesh is quadrilateral. Each mesh cell is denoted by four vertices with one centre point. Vertices
    may be shared with neighbouring cells. The centre points should be unique.

    Raysect's mesh interpolator uses a different mesh scheme. Mesh cells are triangles and data values are stored at the
    triangle vertices. Therefore, each SOLPS quadrilateral cell is split into two triangular cells. The data points are
    later interpolated onto the vertex points.

    :param ndarray r: Array of cell vertex r coordinates, must be 3 dimensional. Example shape is (4 x 32 x 98).
    :param ndarray z: Array of cell vertex z coordinates, must be 3 dimensional. Example shape is (4 x 32 x 98).
    :param ndarray vol: Array of cell volumes in m-3. Example shape is (32 x 98).
    :param ndarray neighbix: Array of poloidal indeces of neighbouring cells in order: left, bottom, right, top,
                             must be 3 dimensional. Example shape is (4 x 32 x 98).
                             In SOLPS notation: left/right - poloidal prev./next, bottom/top - radial prev./next.
                             Cell indexing starts with 0 and -1 means no neighbour.
    :param ndarray neighbiy: Array of radial indeces of neighbouring cells in order: left, bottom, right, top,
                             must be 3 dimensional. Example shape is (4 x 32 x 98).
    """

    # TODO Make neighbix and neighbix optional in the future, as they can be reconstructed with _triangle_to_grid_map

    def __init__(self, r, z, vol, neighbix, neighbiy):

        if r.shape != z.shape:
            raise ValueError('Shape of r array: {0} mismatch the shape of z array: {1}.'.format(r.shape, z.shape))

        if vol.shape != r.shape[1:]:
            raise ValueError('Shape of vol array: {0} mismatch the grid dimentions: {1}.'.format(vol.shape, r.shape[1:]))

        if neighbix.shape != r.shape:
            raise ValueError('Shape of neighbix array must be {0}, but it is  {1}.'.format(r.shape, neighbix.shape))

        if neighbiy.shape != r.shape:
            raise ValueError('Shape of neighbix array must be {0}, but it is  {}.'.format(r.shape, neighbiy.shape))

        self._r = r
        self._z = z

        self._vol = vol

        self._neighbix = neighbix.astype(int)
        self._neighbiy = neighbiy.astype(int)

        self._initial_setup()

    def _initial_setup(self):

        r = self._r
        z = self._z

        self._cr = r.sum(0) / 4.
        self._cz = z.sum(0) / 4.

        self._nx = r.shape[2]  # poloidal
        self._ny = r.shape[1]  # radial

        self.vessel = None

        # Calculating poloidal basis vector
        self._poloidal_basis_vector = np.zeros((2, self._ny, self._nx))
        vec_r = r[1] - r[0]
        vec_z = z[1] - z[0]
        vec_magn = np.sqrt(vec_r**2 + vec_z**2)
        self._poloidal_basis_vector[0] = np.divide(vec_r, vec_magn, out=np.zeros_like(vec_magn), where=(vec_magn > 0))
        self._poloidal_basis_vector[1] = np.divide(vec_z, vec_magn, out=np.zeros_like(vec_magn), where=(vec_magn > 0))

        # Calculating radial contact areas
        self._radial_area = np.pi * (r[1] + r[0]) * vec_magn

        # Calculating radial basis vector
        self._radial_basis_vector = np.zeros((2, self._ny, self._nx))
        vec_r = r[2] - r[0]
        vec_z = z[2] - z[0]
        vec_magn = np.sqrt(vec_r**2 + vec_z**2)
        self._radial_basis_vector[0] = np.divide(vec_r, vec_magn, out=np.zeros_like(vec_magn), where=(vec_magn > 0))
        self._radial_basis_vector[1] = np.divide(vec_z, vec_magn, out=np.zeros_like(vec_magn), where=(vec_magn > 0))

        # Calculating poloidal contact areas
        self._poloidal_area = np.pi * (r[2] + r[0]) * vec_magn

        # For convertion from Cartesian to poloidal
        # TODO Make it work with triangle cells
        self._inv_det = 1. / (self._poloidal_basis_vector[0] * self._radial_basis_vector[1] -
                              self._poloidal_basis_vector[1] * self._radial_basis_vector[0])

        # Finding unique vertices
        vertices = np.array([r.flatten(), z.flatten()]).T
        self._vertex_coords, unique_vertices = np.unique(vertices, axis=0, return_inverse=True)
        self._num_vertices = self._vertex_coords.shape[0]

        # Work out the extent of the mesh.
        self._mesh_extent = {"minr": r.min(), "maxr": r.max(), "minz": z.min(), "maxz": z.max()}

        # Pull out the index number for each unique vertex in this rectangular cell.
        # Unusual vertex indexing is based on SOLPS output, see Matlab code extract from David Moulton.
        ng = self._nx * self._ny  # grid size
        ym, xm = np.meshgrid(np.arange(self._ny, dtype=np.int32), np.arange(self._nx, dtype=np.int32), indexing='ij')

        # add quadrangle b2 grid
        self._quadrangles = np.zeros((ng, 4), dtype=np.int32)
        self._quadrangles[:, 0] = unique_vertices[0:ng]
        self._quadrangles[:, 1] = unique_vertices[2 * ng: 3 * ng]
        self._quadrangles[:, 2] = unique_vertices[3 * ng: 4 * ng]
        self._quadrangles[:, 3] = unique_vertices[ng: 2 * ng]

        # add mapping from quadrangles to the b2 grid
        self._quadrangle_to_grid_map = np.zeros((ng, 2), dtype=np.int32)
        self._quadrangle_to_grid_map[:, 0] = ym.flatten()
        self._quadrangle_to_grid_map[:, 1] = xm.flatten()

        # Number of triangles must be equal to number of rectangle centre points times 2.
        self._num_tris = ng * 2
        self._triangles = np.zeros((self._num_tris, 3), dtype=np.int32)
        self._triangles[0::2, 0] = self._quadrangles[:, 0]
        self._triangles[0::2, 1] = self._quadrangles[:, 1]
        self._triangles[0::2, 2] = self._quadrangles[:, 2]
        # Split the quad cell into two triangular cells.
        self._triangles[1::2, 0] = self._quadrangles[:, 2]
        self._triangles[1::2, 1] = self._quadrangles[:, 3]
        self._triangles[1::2, 2] = self._quadrangles[:, 0]

        # Each triangle cell is mapped to the tuple ID (ix, iy) of its parent mesh cell.
        self._triangle_to_grid_map = np.zeros((self._num_tris, 2), dtype=np.int32)
        self._triangle_to_grid_map[::2, 0] = ym.flatten()
        self._triangle_to_grid_map[::2, 1] = xm.flatten()
        self._triangle_to_grid_map[1::2] = self._triangle_to_grid_map[::2]

    @property
    def nx(self):
        """Number of grid cells in the poloidal direction."""
        return self._nx

    @property
    def ny(self):
        """Number of grid cells in the radial direction."""
        return self._ny

    @property
    def cr(self):
        """R-coordinate of the cell centres."""
        return self._cr

    @property
    def cz(self):
        """Z-coordinate of the cell centres."""
        return self._cz

    @property
    def r(self):
        """R-coordinates of the cell vertices."""
        return self._r

    @property
    def z(self):
        """Z-coordinate of the cell vertices."""
        return self._z

    @property
    def vol(self):
        """Volume of each grid cell in m-3."""
        return self._vol

    @property
    def neighbix(self):
        """Poloidal indeces of neighbouring cells in order: left, bottom, right, top."""
        return self._neighbix

    @property
    def neighbiy(self):
        """Radial indeces of neighbouring cells in order: left, bottom, right, top."""
        return self._neighbiy

    @property
    def radial_area(self):
        """Radial contact area in m-2."""
        return self._radial_area

    @property
    def poloidal_area(self):
        """Poloidal contact area in m-2."""
        return self._poloidal_area

    @property
    def vertex_coordinates(self):
        """RZ-coordinates of unique vertices."""
        return self._vertex_coords

    @property
    def num_vertices(self):
        """Total number of unique vertices."""
        return self._num_vertices

    @property
    def mesh_extent(self):
        """Extent of the mesh. A dictionary with minr, maxr, minz and maxz keys."""
        return self._mesh_extent

    @property
    def num_triangles(self):
        """Total number of triangles (the number of cells doubled)."""
        return self._num_tris

    @property
    def triangles(self):
        """Array of triangle vertex indices with (num_thiangles, 3) shape."""
        return self._triangles

    @property
    def quadrangles(self):
        """Array of quadrangle vertex indices with (num_thiangles, 3) shape."""
        return self._quadrangles

    @property
    def poloidal_basis_vector(self):
        """
        Array of 2D poloidal basis vectors for grid cells.

        For each cell there is a poloidal and radial basis vector.

        Any vector on the poloidal grid can be converted to cartesian with the following transformation.
        bx = (p_x  r_x) ( b_p )
        by   (p_y  r_y) ( b_r )

        :return: ndarray with shape (2, ny, nx).
        """
        return self._poloidal_basis_vector

    @property
    def radial_basis_vector(self):
        """
        Array of 2D radial basis vectors for grid cells.

        For each cell there is a poloidal and radial basis vector.

        Any vector on the poloidal grid can be converted to cartesian with the following transformation.
        bx = (p_x  r_x) ( b_p )
        by   (p_y  r_y) ( b_r )

        :return: ndarray with shape (2, ny, nx).
        """
        return self._radial_basis_vector

    @property
    def triangle_to_grid_map(self):
        """
        Array mapping every triangle index to a tuple grid cell ID, i.e. (iy, ix).

        :return: ndarray with shape (nx*ny*2, 2)
        """
        return self._triangle_to_grid_map

    @property
    def quadrangle_to_grid_map(self):
        """
        Array mapping every quadrangle index to a tuple grid cell ID, i.e. (iy, ix).

        :return: ndarray with shape (nx*ny, 2)
        """
        return self._quadrangle_to_grid_map

    def __getstate__(self):
        state = {
            'r': self._r,
            'z': self._z,
            'vol': self._vol,
            'neighbix': self._neighbix,
            'neighbiy': self._neighbiy
        }
        return state

    def __setstate__(self, state):
        self._r = state['r']
        self._z = state['z']
        self._vol = state['vol']
        self._neighbix = state['neighbix']
        self._neighbiy = state['neighbiy']
        self._initial_setup()

    def to_cartesian(self, vec_pol):
        """
        Converts the 2D vector defined on mesh from poloidal to cartesian coordinates.
        :param ndarray vec_pol: Array of 2D vector with with shape (2, ny, nx).
            [0, :, :] - poloidal component, [1, :, :] - radial component

        :return: ndarray with shape (2, ny, nx)
        """
        vec_cart = np.zeros((2, self._ny, self._nx))
        vec_cart[0] = self._poloidal_basis_vector[0] * vec_pol[0] + self._radial_basis_vector[0] * vec_pol[1]
        vec_cart[1] = self._poloidal_basis_vector[1] * vec_pol[0] + self._radial_basis_vector[1] * vec_pol[1]

        return vec_cart

    def to_poloidal(self, vec_cart):
        """
        Converts the 2D vector defined on mesh from cartesian to poloidal coordinates.
        :param ndarray vector_on_mesh: Array of 2D vector with with shape (2, ny, nx).
            [0, :, :] - R component, [1, :, :] - Z component

        :return: ndarray with shape (2, ny, nx)
        """
        vec_pol = np.zeros((2, self._ny, self._nx))
        vec_pol[0] = self._inv_det * (self._radial_basis_vector[1] * vec_cart[0] - self._radial_basis_vector[0] * vec_cart[1])
        vec_pol[1] = self._inv_det * (self._poloidal_basis_vector[0] * vec_cart[1] - self._poloidal_basis_vector[1] * vec_cart[0])

        return vec_pol

    def plot_triangle_mesh(self, solps_data=None, ax=None):
        """
        Plot the triangle mesh grid geometry to a matplotlib figure.

        :param solps_data: Data array defined on the SOLPS mesh
        """
        if ax is None:
            _, ax = plt.subplots(constrained_layout=True)

        verts = self.vertex_coordinates[self.triangles]
        if solps_data is None:
            collection_mesh = PolyCollection(verts, facecolor="none", edgecolor='b', linewidth=0.5)
        else:
            collection_mesh = PolyCollection(verts)
            collection_mesh.set_array(solps_data[self.triangle_to_grid_map[:, 0], self.triangle_to_grid_map[:, 1]])
        ax.add_collection(collection_mesh)
        ax.set_aspect(1)
        ax.set_xlim(self.mesh_extent["minr"], self.mesh_extent["maxr"])
        ax.set_ylim(self.mesh_extent["minz"], self.mesh_extent["maxz"])
        ax.set_xlabel("R [m]")
        ax.set_ylabel("Z [m]")

        return ax

    def plot_quadrangle_mesh(self, solps_data=None, ax=None):
        """
        Plot the quadrangle mesh grid geometry to a matplotlib figure.

        :param solps_data: Data array defined on the SOLPS mesh
        """

        if ax is None:
            _, ax = plt.subplots(constrained_layout=True)

        verts = self.vertex_coordinates[self.quadrangles]
        if solps_data is None:
            collection_mesh = PolyCollection(verts, facecolor="none", edgecolor='b', linewidth=0.5)
        else:
            collection_mesh = PolyCollection(verts)
            collection_mesh.set_array(solps_data[self.quadrangle_to_grid_map[:, 0], self.quadrangle_to_grid_map[:, 1]])
        ax.add_collection(collection_mesh)
        ax.set_aspect(1)
        ax.set_xlim(self.mesh_extent["minr"], self.mesh_extent["maxr"])
        ax.set_ylim(self.mesh_extent["minz"], self.mesh_extent["maxz"])
        ax.set_xlabel("R [m]")
        ax.set_ylabel("Z [m]")

        return ax
